from enum import StrEnum

__all__ = ("CombatType", "PathType")


class PathType(StrEnum):
    KNIGHT = "Knight"
    """Preservation"""
    ROGUE = "Rogue"
    """The Hunt"""
    MAGE = "Mage"
    """Erudition"""
    WARLOCK = "Warlock"
    """Nihility"""
    WARRIOR = "Warrior"
    """Destruction"""
    SHAMAN = "Shaman"
    """Harmony"""
    PRIEST = "Priest"
    """Abundance"""


class CombatType(StrEnum):
    ICE = "Ice"
    FIRE = "Fire"
    QUANTUM = "Quantum"
    IMAGINARY = "Imaginary"
    WIND = "Wind"
    THUNDER = "Thunder"
    PHYSICAL = "Physical"



import datetime
from typing import Any

from pydantic import Field, field_validator

from ..enums import CombatType, PathType
from ..utils import format_str, replace_placeholders
from .base import BaseModel

__all__ = (
    "BaseSkill",
    "Character",
    "CharacterAscensionItem",
    "CharacterDetail",
    "CharacterDetailTypes",
    "CharacterEidolon",
    "CharacterInfo",
    "CharacterScript",
    "CharacterStory",
    "CharacterTraces",
    "CharacterType",
    "CharacterUpgrade",
    "CharacterVoice",
    "ExtraEffect",
    "SkillAdd",
    "SkillListSkill",
    "SkillPoint",
    "SkillPromote",
    "SkillPromoteCostItem",
    "SkillTree",
    "SkillTreeSkill",
    "Status",
    "VoiceActor",
    "WeaknessBreak",
)


class CharacterStory(BaseModel):
    title: str
    text: str


class CharacterVoice(BaseModel):
    title: str
    text: str
    audio: int | None


class CharacterScript(BaseModel):
    stories: list[CharacterStory] = Field(alias="story")
    voices: list[CharacterVoice] = Field(alias="voice")

    @field_validator("stories", mode="before")
    def _convert_stories(cls, v: list[dict[str, Any]] | None) -> list[CharacterStory]:
        return [CharacterStory(**s) for s in v] if v else []

    @field_validator("voices", mode="before")
    def _convert_voices(cls, v: list[dict[str, Any]] | None) -> list[CharacterVoice]:
        return [CharacterVoice(**s) for s in v] if v else []


class CharacterAscensionItem(BaseModel):
    id: int
    amount: int


class SkillAdd(BaseModel):
    """Skill that inceases its level because of an eidolon"""

    id: int
    """ID of the skill"""
    level: int
    """Level added to the skill"""


class CharacterEidolon(BaseModel):
    id: int
    rank: int
    name: str

    params: list[int | float] | None
    description: str
    skill_add_level_list: list[SkillAdd] = Field(alias="skillAddLevelList")
    """List of skills that increase their level because of this eidolon"""
    icon: str

    @field_validator("description", mode="before")
    def _format_description(cls, v: str, values: Any) -> str:
        params = values.data.get("params")
        return replace_placeholders(format_str(v), params)

    @field_validator("skill_add_level_list", mode="before")
    def _convert_skill_add_level_list(cls, v: dict[str, int] | None) -> list[SkillAdd]:
        return [SkillAdd(id=int(id_), level=level) for id_, level in v.items()] if v else []

    @field_validator("icon", mode="before")
    def _convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/skill/{v}.png"


class SkillPromoteCostItem(BaseModel):
    id: int
    amount: int


class SkillPromote(BaseModel):
    level: int
    cost_items: list[SkillPromoteCostItem] = Field(alias="costItems")

    @field_validator("cost_items", mode="before")
    def _convert_cost_items(cls, v: dict[str, dict[str, int] | None]) -> list[SkillPromoteCostItem]:
        return (
            [SkillPromoteCostItem(id=int(id_), amount=a) for id_, a in v["costItems"].items()]
            if v["costItems"]
            else []
        )


class Status(BaseModel):
    name: str
    value: int | float
    icon: str

    @field_validator("icon", mode="before")
    def _convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/status/{v}.png"


class ExtraEffect(BaseModel):
    name: str
    description: str
    icon: str


class WeaknessBreak(BaseModel):
    type: str
    value: int


class SkillPoint(BaseModel):
    type: str
    value: int | None


class SkillListSkill(BaseModel):
    id: int
    name: str
    tag: str | None
    type: str

    max_level: int = Field(alias="maxLevel")
    skill_points: list[SkillPoint] = Field(alias="skillPoints")
    weakness_break: list[WeaknessBreak] = Field(alias="weaknessBreak")
    description: str | None
    simplified_description: str | None = Field(alias="descriptionSimple")

    traces: list[int]
    eidolons: list[int]
    extra_effects: list[ExtraEffect] = Field(alias="extraEffects")
    attack_type: str | None = Field(alias="attackType")
    damage_type: str | None = Field(alias="damageType")
    icon: str

    params: dict[str, list[float]] | None

    @field_validator("skill_points", mode="before")
    def _convert_skill_points(cls, v: dict[str, int | None]) -> list[SkillPoint]:
        return [SkillPoint(type=k, value=v) for k, v in v.items()]

    @field_validator("weakness_break", mode="before")
    def _convert_weakness_break(cls, v: dict[str, int] | None) -> list[WeaknessBreak]:
        return [WeaknessBreak(type=k, value=v) for k, v in v.items()] if v else []

    @field_validator("simplified_description", mode="before")
    def _format_simplified_description(cls, v: str | None) -> str | None:
        return format_str(v) if v else None

    @field_validator("traces", mode="before")
    def _convert_traces(cls, v: list[int] | None) -> list[int]:
        return v if v else []

    @field_validator("eidolons", mode="before")
    def _convert_eidolons(cls, v: list[int] | None) -> list[int]:
        return v if v else []

    @field_validator("extra_effects", mode="before")
    def _convert_extra_effects(cls, v: list[dict[str, Any]] | None) -> list[ExtraEffect]:
        return [ExtraEffect(**e) for e in v] if v else []

    @field_validator("icon", mode="before")
    def _convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/skill/{v}.png"


class BaseSkill(BaseModel):
    id: int
    name: str | None
    description: str | None

    point_type: str = Field(alias="pointType")
    point_position: str = Field(alias="pointPosition")
    max_level: int = Field(alias="maxLevel")
    is_default: bool = Field(alias="isDefault")

    avatar_level_limit: int | None = Field(alias="avatarLevelLimit")
    avatar_promotion_limit: int | None = Field(alias="avatarPromotionLimit")

    skill_list: list[SkillListSkill] = Field(alias="skillList")
    status_list: list[Status] = Field(alias="statusList")
    icon: str
    params: dict[str, list[float]] | None

    promote: list[SkillPromote]

    @field_validator("skill_list", mode="before")
    def _convert_skill_list(cls, v: dict[str, dict[str, Any]] | None) -> list[SkillListSkill]:
        return [SkillListSkill(id=int(s), **v[s]) for s in v] if v else []

    @field_validator("status_list", mode="before")
    def _convert_status_list(cls, v: list[dict[str, Any]] | None) -> list[Status]:
        return [Status(**s) for s in v] if v else []

    @field_validator("icon", mode="before")
    def _convert_icon(cls, v: str) -> str:
        if "SkillIcon" in v:
            return f"https://api.yatta.top/hsr/assets/UI/skill/{v}.png"
        return f"https://api.yatta.top/hsr/assets/UI/status/{v}.png"

    @field_validator("promote", mode="before")
    def _convert_promote(cls, v: dict[str, dict[str, dict[str, int] | None]]) -> list[SkillPromote]:
        return [SkillPromote(level=int(p), costItems=v[p]) for p in v] if v else []  # type: ignore


class SkillTreeSkill(BaseModel):
    id: int
    points_direction: str | None = Field(alias="pointsDirection")
    points: list[int]

    @field_validator("points", mode="before")
    def _convert_points(cls, v: list[int] | None) -> list[int]:
        return v if v else []


class SkillTree(BaseModel):
    id: int
    type: str
    tree: list[SkillTreeSkill] = Field([])

    @field_validator("tree", mode="before")
    def _convert_tree(cls, v: dict[str, dict[str, Any]]) -> list[SkillTreeSkill]:
        return [SkillTreeSkill(**v[s]) for s in v]


class CharacterTraces(BaseModel):
    main_skills: list[BaseSkill] = Field(alias="mainSkills")
    sub_skills: list[BaseSkill] = Field(alias="subSkills")
    tree_skills: list[SkillTree] = Field(alias="skillsTree")

    @field_validator("main_skills", mode="before")
    def _convert_main_skills(cls, v: dict[str, dict[str, Any]]) -> list[BaseSkill]:
        return [BaseSkill(**v[s]) for s in v]

    @field_validator("sub_skills", mode="before")
    def _convert_sub_skills(cls, v: dict[str, dict[str, Any]]) -> list[BaseSkill]:
        return [BaseSkill(**v[s]) for s in v]

    @field_validator("tree_skills", mode="before")
    def _convert_tree_skills(cls, v: dict[str, dict[str, Any]]) -> list[SkillTree]:
        return [SkillTree(**v[s]) for s in v]


class CharacterCostItem(BaseModel):
    id: int
    amount: int


class CharacterUpgrade(BaseModel):
    level: int
    cost_items: list[CharacterCostItem] = Field(alias="costItems")
    max_level: int = Field(alias="maxLevel")
    required_player_level: int = Field(alias="playerLevelRequire")
    required_world_level: int = Field(alias="worldLevelRequire")
    skill_base: dict[str, int | float] = Field(alias="skillBase")
    skill_add: dict[str, int | float] = Field(alias="skillAdd")

    @field_validator("cost_items", mode="before")
    def _convert_cost_items(cls, v: dict[str, int] | None) -> list[CharacterCostItem]:
        return [CharacterCostItem(id=int(k), amount=v) for k, v in v.items()] if v else []

    @field_validator("required_player_level", mode="before")
    def _convert_required_player_level(cls, v: int | None) -> int:
        return v if v else 0

    @field_validator("required_world_level", mode="before")
    def _convert_required_world_level(cls, v: int | None) -> int:
        return v if v else 0


class VoiceActor(BaseModel):
    lang: str
    name: str


class CharacterInfo(BaseModel):
    faction: str | None
    description: str
    voice_actors: list[VoiceActor] = Field(alias="cv")

    @field_validator("voice_actors", mode="before")
    def _convert_voice_actors(cls, v: dict[str, str] | None) -> list[VoiceActor]:
        return [VoiceActor(lang=k, name=v) for k, v in v.items()] if v else []


class CharacterDetailType(BaseModel):
    id: str
    name: str


class CharacterDetailTypes(BaseModel):
    path_type: CharacterDetailType = Field(alias="pathType")
    combat_type: CharacterDetailType = Field(alias="combatType")


class CharacterDetail(BaseModel):
    id: int
    name: str
    beta: bool = Field(False)
    rarity: int = Field(alias="rank")
    types: CharacterDetailTypes
    icon: str
    release: int
    route: str
    info: CharacterInfo = Field(alias="fetter")
    upgrades: list[CharacterUpgrade] = Field(alias="upgrade")
    traces: CharacterTraces
    eidolons: list[CharacterEidolon]
    ascension: list[CharacterAscensionItem]
    script: CharacterScript
    release_at: datetime.datetime | None = Field(None, alias="release")

    @field_validator("icon", mode="before")
    def _convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/avatar/{v}.png"

    @field_validator("eidolons", mode="before")
    def _convert_eidolons(cls, v: dict[str, dict[str, Any]]) -> list[CharacterEidolon]:
        return [CharacterEidolon(**v[s]) for s in v]

    @field_validator("ascension", mode="before")
    def _convert_ascension(cls, v: dict[str, int]) -> list[CharacterAscensionItem]:
        return [CharacterAscensionItem(id=int(k), amount=v) for k, v in v.items()]

    @field_validator("release_at", mode="before")
    def _convert_release_at(cls, v: int | None) -> datetime.datetime | None:
        return datetime.datetime.fromtimestamp(v) if v else None

    @property
    def medium_icon(self) -> str:
        return self.icon.replace("avatar", "avatar/medium")

    @property
    def large_icon(self) -> str:
        return self.icon.replace("avatar", "avatar/large")

    @property
    def round_icon(self) -> str:
        return self.icon.replace("avatar", "avatar/round")


class CharacterType(BaseModel):
    path_type: PathType = Field(alias="pathType")
    combat_type: CombatType = Field(alias="combatType")


class Character(BaseModel):
    id: int
    name: str
    rarity: int = Field(alias="rank")
    icon: str
    types: CharacterType
    route: str
    beta: bool = Field(False)
    release_at: datetime.datetime | None = Field(None, alias="release")

    @field_validator("icon", mode="before")
    def _convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/avatar/{v}.png"

    @field_validator("release_at", mode="before")
    def _convert_release_at(cls, v: int | None) -> datetime.datetime | None:
        return datetime.datetime.fromtimestamp(v) if v else None

    @property
    def medium_icon(self) -> str:
        return self.icon.replace("avatar", "avatar/medium")

    @property
    def large_icon(self) -> str:
        return self.icon.replace("avatar", "avatar/large")

    @property
    def round_icon(self) -> str:
        return self.icon.replace("avatar", "avatar/round")



from typing import Any

from pydantic import Field, field_validator

from .base import BaseModel

__all__ = (
    "Book",
    "BookDetail",
    "BookSeries",
)


class BookSeries(BaseModel):
    """Represents a book series.

    Attributes:
        id (int): The ID of the series.
        name (str): The name of the series.
        story (str): The story of the series.
        image_list (list[str]): A list of image URLs.
    """

    id: int
    name: str
    story: str
    image_list: list[str] = Field(alias="imageList")

    @field_validator("image_list", mode="before")
    def _convert_image_list(cls, v: list[str] | None) -> list[str]:
        return v if v else []


class BookDetail(BaseModel):
    """Represents a book.

    Attributes:
        id (int): The ID of the book.
        name (str): The name of the book.
        world_type (str): The type of world the book is in.
        chapter_count (int): The number of chapters in the book.
        icon (str): The URL of the book's icon.
        description (str): The description of the book.
        series (list[BookSeries]): A list of book series.
    """

    id: int
    name: str
    world_type: str = Field(alias="worldType")
    chapter_count: int = Field(0)
    icon: str
    description: str
    series: list[BookSeries]

    @field_validator("icon", mode="before")
    def _convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/item/{v}.png"

    @field_validator("series", mode="before")
    def _convert_series(cls, v: dict[str, dict[str, Any]]) -> list[BookSeries]:
        return [BookSeries(id=int(series_id), **s) for series_id, s in v.items()]


class Book(BaseModel):
    """Represents a book.

    Attributes:
        id (int): The ID of the book.
        name (str): The name of the book.
        world_type (int): The type of world the book is in.
        chapter_count (int): The number of chapters in the book.
        icon (str): The URL of the book's icon.
        route (str): The route of the book.
    """

    id: int
    name: str
    world_type: int = Field(alias="worldType")
    chapter_count: int = Field(alias="chapterCount")
    icon: str
    route: str

    @field_validator("icon", mode="before")
    def _convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/item/{v}.png"



from pydantic import Field, field_validator

from .base import BaseModel

__all__ = ("Contact", "Message")


class Contact(BaseModel):
    name: str
    signature: str | None = Field(None)
    type: int
    icon: str

    @field_validator("icon", mode="before")
    def convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/avatar/{v}.png"


class Message(BaseModel):
    id: int
    contact: Contact = Field(alias="contacts")
    section_count: int = Field(alias="sectionCount")
    route: str | None = None



from typing import Any

from pydantic import Field, field_validator

from ..utils import replace_placeholders
from .base import BaseModel

__all__ = (
    "Relic",
    "RelicSet",
    "RelicSetDetail",
)


class Relic(BaseModel):
    pos: str
    name: str
    description: str
    story: str
    icon: str

    @field_validator("icon", mode="before")
    def _convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/relic/{v}.png"


class SetEffect(BaseModel):
    params: dict[str, list[int | float]] | None
    description: str

    @field_validator("description", mode="before")
    def _format_description(cls, v: str, values: Any) -> str:
        params = values.data.get("params")
        return replace_placeholders(v, params)


class SetEffects(BaseModel):
    two_piece: SetEffect = Field(alias="2")
    four_piece: SetEffect | None = Field(None, alias="4")


class RelicSetDetail(BaseModel):
    id: int
    name: str
    icon: str
    rarity_list: list[int] = Field(alias="levelList")
    is_planar_suit: bool = Field(alias="isPlanarSuit")
    route: str
    beta: bool = Field(False)

    set_effects: SetEffects = Field(alias="skillList")
    relics: list[Relic] = Field(alias="suite")

    @field_validator("icon", mode="before")
    def convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/relic/{v}.png"

    @field_validator("relics", mode="before")
    def convert_relics(cls, v: dict[str, dict[str, Any]]) -> list[Relic]:
        return [Relic(pos=pos, **v[pos]) for pos in v]


class RelicSet(BaseModel):
    id: int
    name: str
    beta: bool = Field(False)
    icon: str
    rarity_list: list[int] = Field(alias="levelList")
    is_planar_suit: bool = Field(alias="isPlanarSuit")
    route: str

    @field_validator("icon", mode="before")
    def convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/relic/{v}.png"



from typing import Any

from pydantic import Field, field_validator

from .base import BaseModel

__all__ = (
    "Item",
    "ItemDetail",
    "ItemSource",
    "ItemType",
    "Recipe",
    "RecipeMaterial",
)


class RecipeMaterial(BaseModel):
    id: int
    rarity: int = Field(alias="rank")
    icon: str
    amount: int = Field(alias="count")

    @field_validator("icon", mode="before")
    def _convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/item/{v}.png"


class Recipe(BaseModel):
    coin_cost: int = Field(alias="coinCost")
    required_world_level: int = Field(alias="worldLevelRequire")
    materials: list[RecipeMaterial] = Field(alias="materialCost")
    special_materials: list[RecipeMaterial] = Field(alias="specialMaterialCost")

    @field_validator("coin_cost", mode="before")
    def _convert_coin_cost(cls, v: int | None) -> int:
        return v or 0

    @field_validator("required_world_level", mode="before")
    def _convert_required_world_level(cls, v: int | None) -> int:
        return v or 0

    @field_validator("materials", mode="before")
    def _convert_materials(cls, v: dict[str, dict[str, Any]] | None) -> list[RecipeMaterial]:
        return [RecipeMaterial(id=int(id_), **m) for id_, m in v.items()] if v else []

    @field_validator("special_materials", mode="before")
    def _convert_special_materials(
        cls, v: dict[str, dict[str, Any]] | None
    ) -> list[RecipeMaterial]:
        return [RecipeMaterial(id=int(id_), **m) for id_, m in v.items()] if v else []


class ItemSource(BaseModel):
    description: str
    recipes: list[Recipe] = Field(alias="recipe")

    @field_validator("recipes", mode="before")
    def _convert_recipes(cls, v: list[dict[str, Any]] | None) -> list[Recipe]:
        return [Recipe(**r) for r in v] if v else []


class ItemType(BaseModel):
    id: int
    name: str


class ItemDetail(BaseModel):
    id: int
    name: str
    beta: bool = Field(False)
    rarity: int = Field(alias="rank")
    tags: list[str]
    icon: str
    route: str
    description: str
    story: str | None
    sources: list[ItemSource] = Field(alias="source")

    @field_validator("icon", mode="before")
    def _convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/item/{v}.png"

    @field_validator("sources", mode="before")
    def _convert_sources(cls, v: list[dict[str, Any]]) -> list[ItemSource]:
        return [ItemSource(**s) for s in v] if v else []


class Item(BaseModel):
    id: int
    name: str
    beta: bool = Field(False)
    rarity: int = Field(alias="rank")
    type: int
    tags: list[str]
    icon: str
    route: str

    @field_validator("icon", mode="before")
    def _convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/item/{v}.png"



from __future__ import annotations

import logging
from enum import Enum
from typing import TYPE_CHECKING, Any, Final, Self

from aiohttp_client_cache.backends.sqlite import SQLiteBackend
from aiohttp_client_cache.session import CachedSession

from .exceptions import ConnectionTimeoutError, DataNotFoundError, YattaAPIError
from .models import (
    Book,
    BookDetail,
    Changelog,
    Character,
    CharacterDetail,
    Item,
    ItemDetail,
    LightCone,
    LightConeDetail,
    Message,
    RelicSet,
    RelicSetDetail,
)

if TYPE_CHECKING:
    import aiohttp

__all__ = ("Language", "YattaAPI")

LOGGER_ = logging.getLogger("yatta.py")


class Language(Enum):
    CHT = "cht"
    CN = "cn"
    DE = "de"
    EN = "en"
    ES = "es"
    FR = "fr"
    ID = "id"
    JP = "jp"
    KR = "kr"
    PT = "pt"
    RU = "ru"
    TH = "th"
    VI = "vi"


class YattaAPI:
    """
    The main class that is used to interact with the API.

    Parameters:
        lang: The language to use for the API. Defaults to Language.EN.
        cache_ttl: The time-to-live for the cache in seconds. Defaults to 3600.
        headers: A dictionary of headers to use for the requests. Defaults to None.
        session: An aiohttp.ClientSession to use for the requests. Defaults to None.
    """

    BASE_URL: Final[str] = "https://api.yatta.top/hsr/v2"

    def __init__(
        self,
        *,
        lang: Language = Language.EN,
        cache_ttl: int = 3600,
        headers: dict[str, Any] | None = None,
        session: aiohttp.ClientSession | None = None,
    ) -> None:
        self.lang = lang
        self.cache_ttl = cache_ttl

        self._session = session
        self._cache = SQLiteBackend("./.cache/yatta/aiohttp-cache.db", expire_after=cache_ttl)
        self._headers = headers or {"User-Agent": "yatta-py"}

    async def __aenter__(self) -> Self:
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:  # noqa: ANN001
        await self.close()

    async def _request(
        self, endpoint: str, *, static: bool = False, use_cache: bool
    ) -> dict[str, Any]:
        """
        A helper function to make requests to the API.

        Parameters
        ----------
        endpoint : str
            The endpoint to request from.
        static : bool, optional
            Whether to use the static endpoint or not. Defaults to False.
        use_cache : bool
            Whether to use the cache or not

        Returns
        -------
        Dict[str, Any]
            The response from the API.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        if self._session is None:
            msg = "Call `start` before making requests."
            raise RuntimeError(msg)

        if static:
            url = f"{self.BASE_URL}/static/{endpoint}"
        else:
            url = f"{self.BASE_URL}/{self.lang.value}/{endpoint}"

        LOGGER_.debug("Requesting %s...", url)

        if not use_cache and isinstance(self._session, CachedSession):
            async with self._session.disabled(), self._session.get(url) as resp:
                if resp.status != 200:
                    self._handle_error(resp.status)
                data: dict[str, Any] = await resp.json()
        else:
            async with self._session.get(url) as resp:
                if resp.status != 200:
                    self._handle_error(resp.status)
                data: dict[str, Any] = await resp.json()

        return data

    def _handle_error(self, code: int) -> None:
        """
        A helper function to handle errors.
        """
        match code:
            case 404:
                raise DataNotFoundError
            case 522:
                raise ConnectionTimeoutError
            case _:
                raise YattaAPIError(code)

    async def start(self) -> None:
        """
        Starts the client session.
        """
        self._session = self._session or CachedSession(headers=self._headers, cache=self._cache)

    async def close(self) -> None:
        """
        Closes the client session and cache.
        """
        if self._session is not None:
            await self._session.close()

    async def fetch_books(self, use_cache: bool = True) -> list[Book]:
        """
        Fetch all books from the API.

        Parameters
        ----------
        use_cache : bool, optional
            Whether to use the cache or not. Defaults to True.

        Returns
        -------
        List[Book]
            A list of Book objects.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        data = await self._request("book", use_cache=use_cache)
        books = [Book(**b) for b in data["data"]["items"].values()]
        return books

    async def fetch_book_detail(self, id: int, use_cache: bool = True) -> BookDetail:
        """
        Fetch a book's detail from the API.

        Parameters
        ----------
        id : int
            The ID of the book to fetch.
        use_cache : bool, optional
            Whether to use the cache or not. Defaults to True.

        Returns
        -------
        BookDetail
            A BookDetail object.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        data = await self._request(f"book/{id}", use_cache=use_cache)
        book = BookDetail(**data["data"])
        return book

    async def fetch_characters(self, use_cache: bool = True) -> list[Character]:
        """
        Fetch all characters from the API.

        Parameters
        ----------
        use_cache : bool, optional
            Whether to use the cache or not. Defaults to True.

        Returns
        -------
        List[Character]
            A list of Character objects.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        data = await self._request("avatar", use_cache=use_cache)
        characters = [Character(**c) for c in data["data"]["items"].values()]
        return characters

    async def fetch_character_detail(self, id: int, use_cache: bool = True) -> CharacterDetail:
        """
        Fetch a character's detail from the API.

        Parameters
        ----------
        id : int
            The ID of the character to fetch.
        use_cache : bool, optional
            Whether to use the cache or not. Defaults to True.

        Returns
        -------
        CharacterDetail
            A CharacterDetail object.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        data = await self._request(f"avatar/{id}", use_cache=use_cache)
        character = CharacterDetail(**data["data"])
        return character

    async def fetch_items(self, use_cache: bool = True) -> list[Item]:
        """
        Fetch all items from the API.

        Parameters
        ----------
        use_cache : bool, optional
            Whether to use the cache or not. Defaults to True.

        Returns
        -------
        List[Item]
            A list of Item objects.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        data = await self._request("item", use_cache=use_cache)
        items = [Item(**i) for i in data["data"]["items"].values()]
        return items

    async def fetch_item_detail(self, id: int, use_cache: bool = True) -> ItemDetail:
        """
        Fetch an item's detail from the API.

        Parameters
        ----------
        id : int
            The ID of the item to fetch.
        use_cache : bool, optional
            Whether to use the cache or not. Defaults to True.

        Returns
        -------
        ItemDetail
            An ItemDetail object.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        data = await self._request(f"item/{id}", use_cache=use_cache)
        item = ItemDetail(**data["data"])
        return item

    async def fetch_light_cones(self, use_cache: bool = True) -> list[LightCone]:
        """
        Fetch all light cones from the API.

        Parameters
        ----------
        use_cache : bool, optional
            Whether to use the cache or not. Defaults to True.

        Returns
        -------
        List[LightCone]
            A list of LightCone objects.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        data = await self._request("equipment", use_cache=use_cache)
        light_cones = [LightCone(**lc) for lc in data["data"]["items"].values()]
        return light_cones

    async def fetch_light_cone_detail(self, id: int, use_cache: bool = True) -> LightConeDetail:
        """
        Fetch a light cone's detail from the API.

        Parameters
        ----------
        id : int
            The ID of the light cone to fetch.
        use_cache : bool, optional
            Whether to use the cache or not. Defaults to True.

        Returns
        -------
        LightConeDetail
            A LightConeDetail object.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        data = await self._request(f"equipment/{id}", use_cache=use_cache)
        light_cone = LightConeDetail(**data["data"])
        return light_cone

    async def fetch_messages(self, use_cache: bool = True) -> list[Message]:
        """
        Fetch all messages from the API.

        Parameters
        ----------
        use_cache : bool, optional
            Whether to use the cache or not. Defaults to True.

        Returns
        -------
        List[Message]
            A list of Message objects.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        data = await self._request("message", use_cache=use_cache)
        messages = [Message(**m) for m in data["data"]["items"].values()]
        return messages

    async def fetch_message_types(self, use_cache: bool = True) -> dict[str, str]:
        """
        Fetch all message types from the API.

        Parameters
        ----------
        use_cache : bool, optional
            Whether to use the cache or not. Defaults to True.

        Returns
        -------
        List[str]
            A list of message types.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        data = await self._request("message", use_cache=use_cache)
        return data["data"]["types"]

    async def fetch_relic_sets(self, use_cache: bool = True) -> list[RelicSet]:
        """
        Fetch all relic sets from the API.

        Parameters
        ----------
        use_cache : bool, optional
            Whether to use the cache or not. Defaults to True.

        Returns
        -------
        List[RelicSet]
            A list of RelicSet objects.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        data = await self._request("relic", use_cache=use_cache)
        relics = [RelicSet(**r) for r in data["data"]["items"].values()]
        return relics

    async def fetch_relic_set_detail(self, id: int, use_cache: bool = True) -> RelicSetDetail:
        """
        Fetch a relic set's detail from the API.

        Parameters
        ----------
        id : int
            The ID of the relic set to fetch.
        use_cache : bool, optional
            Whether to use the cache or not. Defaults to True.

        Returns
        -------
        RelicSetDetail
            A RelicSetDetail object.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        data = await self._request(f"relic/{id}", use_cache=use_cache)
        relic = RelicSetDetail(**data["data"])
        return relic

    async def fetch_changelogs(self, use_cache: bool = True) -> list[Changelog]:
        """
        Fetch changelogs from the API.

        Parameters
        ----------
        use_cache : bool, optional
            Whether to use the cache or not. Defaults to True.

        Returns
        -------
        List[Changelog]
            A list of Changelog objects.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        data = await self._request("changelog", static=True, use_cache=use_cache)
        change_logs: list[Changelog] = []
        for changelog_id, log in data["data"].items():
            change_logs.append(Changelog(id=int(changelog_id), **log))
        return change_logs

    async def fetch_manual_avatar(self, use_cache: bool = True) -> dict[str, dict[str, str]]:
        """
        Fetch the manual avatar from the API.

        Parameters
        ----------
        use_cache : bool, optional
            Whether to use the cache or not. Defaults to True.

        Returns
        -------
        Dict[str, Dict[str, str]]
            A dictionary of avatar stat keys to their corresponding names and icons.

        Raises
        ------
        DataNotFound
            If the requested data is not found.
        """
        data = await self._request("manualAvatar", use_cache=use_cache)
        return data["data"]
